from dataclasses import dataclass
from typing import Callable, Iterable

from elasticai.creator.vhdl.components.fp_hard_sigmoid_componet import (
    FPHardSigmoidComponent,
)
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.vhdl_component import VHDLComponent, VHDLModule


@dataclass
class FPHardSigmoidTranslationArgs:
    fixed_point_factory: Callable[[float], FixedPoint]


@dataclass
class FPHardSigmoidModule(VHDLModule):
    def components(self, args: FPHardSigmoidTranslationArgs) -> Iterable[VHDLComponent]:
        yield FPHardSigmoidComponent(
            zero_threshold=args.fixed_point_factory(-3),
            one_threshold=args.fixed_point_factory(3),
            slope=args.fixed_point_factory(0.125),
            y_intercept=args.fixed_point_factory(0.5),
            fixed_point_factory=args.fixed_point_factory,
        )



from collections.abc import Iterable, Iterator
from dataclasses import dataclass, field
from itertools import chain
from typing import Callable

import numpy as np

from elasticai.creator.vhdl.components import FPLinear1dComponent, RomComponent
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.vhdl_component import VHDLComponent, VHDLModule


@dataclass
class FPLinear1dTranslationArgs:
    fixed_point_factory: Callable[[float], FixedPoint]
    work_library_name: str = field(default="work")


@dataclass
class FPLinear1dModule(VHDLModule):
    layer_id: str
    weight: list[list[float]]
    bias: list[float]

    def components(self, args: FPLinear1dTranslationArgs) -> Iterator[VHDLComponent]:
        def to_fp(values: Iterable[float]) -> list[FixedPoint]:
            return list(map(args.fixed_point_factory, values))

        out_features, in_features = np.shape(self.weight)

        yield FPLinear1dComponent(
            layer_id=self.layer_id,
            in_features=in_features,
            out_features=out_features,
            fixed_point_factory=args.fixed_point_factory,
            work_library_name=args.work_library_name,
            resource_option="auto",
        )

        flat_weight = chain(*self.weight)

        name_suffix = f"_fp_linear_1d_{self.layer_id}"
        yield RomComponent(
            rom_name="w_rom" + name_suffix,
            values=to_fp(flat_weight),
            resource_option="auto",
        )

        yield RomComponent(
            rom_name="b_rom" + name_suffix,
            values=to_fp(self.bias),
            resource_option="auto",
        )



from elasticai.creator.vhdl.quantized_modules.hard_sigmoid import FixedPointHardSigmoid
from elasticai.creator.vhdl.quantized_modules.linear import FixedPointLinear
from elasticai.creator.vhdl.quantized_modules.relu import FixedPointReLU



from elasticai.creator.vhdl.quantized_modules.relu import FixedPointReLU
from elasticai.creator.vhdl.translator.abstract.layers.fp_relu_module import (
    FPReLUModule,
)


def build_fp_relu(fp_relu: FixedPointReLU, layer_id: str) -> FPReLUModule:
    return FPReLUModule()



from elasticai.creator.vhdl.translator.abstract.layers.fp_linear_1d_module import (
    FPLinear1dModule,
    FPLinear1dTranslationArgs,
)
from elasticai.creator.vhdl.translator.abstract.layers.lstm_module import (
    LSTMModule,
    LSTMTranslationArgs,
)



import unittest

from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.quantized_modules.relu import FixedPointReLU
from elasticai.creator.vhdl.translator.abstract.layers.fp_relu_module import (
    FPReLUModule,
)
from elasticai.creator.vhdl.translator.pytorch.build_functions.fp_relu_build_function import (
    build_fp_relu,
)


class FPReluBuildFunctionTest(unittest.TestCase):
    def test_build_function_returns_correct_type(self) -> None:
        layer = FixedPointReLU(
            fixed_point_factory=FixedPoint.get_factory(total_bits=8, frac_bits=4)
        )
        layer_module = build_fp_relu(layer, layer_id="relu1")
        self.assertEqual(type(layer_module), FPReLUModule)



from typing import Any, Callable

import torch

from elasticai.creator.vhdl.evaluators.evaluator import Evaluator


class MetricEvaluator(Evaluator):
    def __init__(
        self,
        inference_evaluator: Evaluator,
        metric: Callable[[torch.Tensor, torch.Tensor], torch.Tensor | float],
    ) -> None:
        self.inference_evaluator = inference_evaluator
        self.metric = metric

    def run(self) -> Any:
        predictions, labels = self.inference_evaluator.run()
        return self.metric(predictions, labels)



import unittest

from elasticai.creator.vhdl.components.fp_linear_1d_component import FPLinear1dComponent
from elasticai.creator.vhdl.number_representations import FixedPoint


class FPLinear1dComponentTest(unittest.TestCase):
    def test_linear_1d_correct_number_of_lines(self) -> None:
        to_fp = FixedPoint.get_factory(total_bits=8, frac_bits=4)

        component = FPLinear1dComponent(
            layer_id="ll1",
            in_features=3,
            out_features=2,
            fixed_point_factory=to_fp,
            work_library_name="work",
        )

        self.assertEqual(len(list(component())), 204)



import unittest

from elasticai.creator.vhdl.components import FPLinear1dComponent, RomComponent
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.abstract.layers import (
    FPLinear1dModule,
    FPLinear1dTranslationArgs,
)


class FPLinear1dModuleTest(unittest.TestCase):
    def setUp(self) -> None:
        self.linear = FPLinear1dModule(
            layer_id="ll1", weight=[[1.0, 2.0, 3.0, 4.0, 5.0, 6.0]], bias=[1.0, 2.0]
        )
        self.translation_args = FPLinear1dTranslationArgs(
            fixed_point_factory=FixedPoint.get_factory(total_bits=8, frac_bits=4)
        )

    def test_contains_all_needed_components(self) -> None:
        vhdl_components = self.linear.components(self.translation_args)

        target_components = [
            (
                FPLinear1dComponent,
                "fp_linear_1d_{layer_name}.vhd".format(layer_name=self.linear.layer_id),
            ),
            (
                RomComponent,
                "w_rom_fp_linear_1d_{layer_name}.vhd".format(
                    layer_name=self.linear.layer_id
                ),
            ),
            (
                RomComponent,
                "b_rom_fp_linear_1d_{layer_name}.vhd".format(
                    layer_name=self.linear.layer_id
                ),
            ),
        ]
        actual_components = [(type(x), x.file_name) for x in vhdl_components]

        self.assertEqual(actual_components, target_components)



from dataclasses import dataclass
from typing import Callable, Iterable

from elasticai.creator.vhdl.components.fp_relu_component import FPReLUComponent
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.vhdl_component import VHDLComponent, VHDLModule


@dataclass
class FPReLUTranslationArgs:
    fixed_point_factory: Callable[[float], FixedPoint]


@dataclass
class FPReLUModule(VHDLModule):
    def components(self, args: FPReLUTranslationArgs) -> Iterable[VHDLComponent]:
        yield FPReLUComponent(
            fixed_point_factory=args.fixed_point_factory,
        )



from unittest import TestCase

import torch.nn

from elasticai.creator.vhdl.translator.pytorch.build_functions import build_lstm


def arange_parameter(
    start: int, end: int, shape: tuple[int, ...]
) -> torch.nn.Parameter:
    return torch.nn.Parameter(
        torch.reshape(torch.arange(start, end, dtype=torch.float32), shape)
    )


class LSTMBuildFunctionTest(TestCase):
    def setUp(self) -> None:
        self.lstm = torch.nn.LSTM(input_size=1, hidden_size=1, num_layers=1)
        self.lstm.weight_ih_l0 = arange_parameter(start=0, end=4, shape=(4, 1))
        self.lstm.weight_hh_l0 = arange_parameter(start=4, end=8, shape=(4, 1))
        self.lstm.bias_ih_l0 = arange_parameter(start=8, end=12, shape=(4,))
        self.lstm.bias_hh_l0 = arange_parameter(start=12, end=16, shape=(4,))

    def test_build_lstm_layer_weights_correct_set(self) -> None:
        lstm_module = build_lstm(self.lstm, layer_id="lstm1")

        self.assertEqual(lstm_module.weights_ih, [[[0.0], [1.0], [2.0], [3.0]]])
        self.assertEqual(lstm_module.weights_hh, [[[4.0], [5.0], [6.0], [7.0]]])

        self.assertEqual(lstm_module.biases_ih, [[8.0, 9.0, 10.0, 11.0]])
        self.assertEqual(lstm_module.biases_hh, [[12.0, 13.0, 14.0, 15.0]])



import unittest

from elasticai.creator.vhdl.components.fp_hard_sigmoid_componet import (
    FPHardSigmoidComponent,
)
from elasticai.creator.vhdl.number_representations import FixedPoint


class FPHardSigmoidComponentTest(unittest.TestCase):
    def test_hard_sigmoid_correct_number_of_lines(self) -> None:
        to_fp = FixedPoint.get_factory(total_bits=8, frac_bits=4)

        component = FPHardSigmoidComponent(
            zero_threshold=to_fp(-3),
            one_threshold=to_fp(3),
            slope=to_fp(0.15),
            y_intercept=to_fp(0.5),
            fixed_point_factory=to_fp,
        )

        self.assertEqual(len(list(component())), 88)



from dataclasses import dataclass, field
from importlib.resources import read_text
from typing import Callable

from elasticai.creator.vhdl.components.utils import (
    calculate_addr_width,
    derive_fixed_point_params_from_factory,
)
from elasticai.creator.vhdl.language import Code
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.templates.utils import expand_template


@dataclass
class FPLinear1dComponent:
    layer_id: str  # used to distinguish layers in the same model
    in_features: int
    out_features: int
    fixed_point_factory: Callable[[float], FixedPoint]
    work_library_name: str = field(default="work")
    resource_option: str = "auto"

    def __post_init__(self) -> None:
        self.data_width, self.frac_width = derive_fixed_point_params_from_factory(
            self.fixed_point_factory
        )
        self.x_addr_width = calculate_addr_width(self.in_features)
        self.y_addr_width = calculate_addr_width(self.out_features)

    @property
    def file_name(self) -> str:
        return f"fp_linear_1d_{self.layer_id}.vhd"

    def __call__(self) -> Code:
        template = read_text("elasticai.creator.vhdl.templates", "fp_linear_1d.tpl.vhd")

        code = expand_template(
            template.splitlines(),
            layer_name=self.layer_id,
            work_library_name=self.work_library_name,
            data_width=self.data_width,
            frac_width=self.frac_width,
            x_addr_width=self.x_addr_width,
            y_addr_width=self.y_addr_width,
            in_feature_num=self.in_features,
            out_feature_num=self.out_features,
            resource_option=f'"{self.resource_option}"',
        )

        yield from code



import unittest

import torch
import torch.nn.functional as F

from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    float_values_to_fixed_point,
)
from elasticai.creator.vhdl.quantized_modules.hard_sigmoid import (
    FixedPointHardSigmoid,
    _HardSigmoidBase,
)


def to_list(x: torch.Tensor) -> list:
    return x.detach().numpy().tolist()


def to_fixed_point(values: list[float], total_bits: int, frac_bits: int) -> list[int]:
    def to_fp(value: FixedPoint) -> int:
        return value.to_signed_int()

    return list(map(to_fp, float_values_to_fixed_point(values, total_bits, frac_bits)))


def from_fixed_point(values: list[int], total_bits: int, frac_bits: int) -> list[float]:
    def to_float(value: int) -> float:
        return float(FixedPoint.from_signed_int(value, total_bits, frac_bits))

    return list(map(to_float, values))


class HardSigmoidBaseTest(unittest.TestCase):
    def test_hard_sigmoid_base_behaves_equal_to_hard_sigmoid(self) -> None:
        hard_sigmoid = _HardSigmoidBase()

        xs = torch.linspace(-6, 6, 100)
        expected = to_list(F.hardsigmoid(xs))
        actual = to_list(hard_sigmoid(xs))

        self.assertEqual(expected, actual)

    def test_hard_sigmoid_base_input_qunat_dequant(self) -> None:
        hard_sigmoid = _HardSigmoidBase(
            input_quant=lambda x: x * 4,
            input_dequant=lambda x: x / 2,
        )

        xs = torch.tensor([-1, 0, 1], dtype=torch.float32)
        expected = to_list(F.hardsigmoid(xs * 2))
        actual = to_list(hard_sigmoid(xs))

        self.assertEqual(expected, actual)

    def test_hard_sigmoid_base_quantized_forward_raises_error(self) -> None:
        with self.assertRaises(NotImplementedError):
            _HardSigmoidBase().quantized_forward(torch.ones(10))


class FixedPointHardSigmoidTest(unittest.TestCase):
    def test_fixed_point_hard_sigmoid_calculates_output_correctly(self) -> None:
        fp_args = 8, 3
        hard_sigmoid = FixedPointHardSigmoid(
            fixed_point_factory=FixedPoint.get_factory(*fp_args)
        )

        xs = torch.linspace(-6, 6, 100)
        quantized_xs = torch.tensor(
            from_fixed_point(to_fixed_point(to_list(xs), *fp_args), *fp_args),
            dtype=torch.float32,
        )
        expected = to_list(F.hardsigmoid(quantized_xs))
        actual = to_list(hard_sigmoid(xs))

        self.assertEqual(expected, actual)

    def test_fixed_point_hard_sigmoid_quantized_forward(self) -> None:
        fp_args = 8, 3
        hard_sigmoid = FixedPointHardSigmoid(
            fixed_point_factory=FixedPoint.get_factory(*fp_args)
        )

        xs = [-3, -1.5, 0, 1.5, 3]
        expected = [0, 3, 4, 5, 8]

        quantized_xs = torch.tensor(to_fixed_point(xs, *fp_args), dtype=torch.float32)
        actual = to_list(hard_sigmoid.quantized_forward(quantized_xs))

        self.assertEqual(expected, actual)



import torch

from elasticai.creator.vhdl.quantized_modules.linear import FixedPointLinear
from elasticai.creator.vhdl.translator.abstract.layers import FPLinear1dModule


def build_fp_linear_1d(linear: FixedPointLinear, layer_id: str) -> FPLinear1dModule:
    def to_list(tensor: torch.Tensor) -> list:
        return tensor.detach().numpy().tolist()

    return FPLinear1dModule(
        layer_id=layer_id,
        weight=to_list(linear.weight),
        bias=to_list(linear.bias),
    )



import unittest

from elasticai.creator.vhdl.components import LSTMComponent
from elasticai.creator.vhdl.number_representations import FixedPoint


class LSTMComponentTest(unittest.TestCase):
    def setUp(self) -> None:
        self.lstm = LSTMComponent(
            input_size=5,
            hidden_size=3,
            fixed_point_factory=FixedPoint.get_factory(total_bits=8, frac_bits=4),
            work_library_name="xil_defaultlib",
        )

    def test_fixed_point_params_correct_derived(self):
        self.assertEqual(8, self.lstm.data_width)
        self.assertEqual(4, self.lstm.frac_width)

    def test_x_h_addr_width_correct_set(self):
        self.assertEqual(3, self.lstm.x_h_addr_width)

    def test_hidden_addr_width_correct_set(self):
        self.assertEqual(3, self.lstm.hidden_addr_width)

    def test_w_addr_width_correct_set(self):
        self.assertEqual(5, self.lstm.w_addr_width)



from collections.abc import Sequence
from dataclasses import dataclass

from elasticai.creator.resource_utils import read_text
from elasticai.creator.vhdl.components.utils import calculate_addr_width, pad_with_zeros
from elasticai.creator.vhdl.language import Code, hex_representation
from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    infer_total_and_frac_bits,
)
from elasticai.creator.vhdl.templates.utils import expand_template


@dataclass
class RomComponent:
    rom_name: str
    values: Sequence[FixedPoint]
    resource_option: str

    def __post_init__(self) -> None:
        self.data_width, _ = infer_total_and_frac_bits(self.values)
        self.addr_width = calculate_addr_width(len(self.values))
        padded_values = pad_with_zeros(list(self.values), 2**self.addr_width)
        self.hex_values = list(
            map(lambda fp: hex_representation(fp.to_hex()), padded_values)
        )

    @property
    def file_name(self) -> str:
        return f"{self.rom_name}.vhd"

    def __call__(self) -> Code:
        template = read_text("elasticai.creator.vhdl.templates", "rom.tpl.vhd")

        code = expand_template(
            template.splitlines(),
            rom_name=self.rom_name,
            rom_addr_bitwidth=self.addr_width,
            rom_data_bitwidth=self.data_width,
            rom_value=",".join(self.hex_values),
            rom_resource_option=f'"{self.resource_option}"',
        )

        yield from code



from elasticai.creator.vhdl.vhdl_component import VHDLStaticComponent


class DualPort2ClockRamComponent(VHDLStaticComponent):
    def __init__(self):
        super().__init__(
            template_package="elasticai.creator.vhdl.templates",
            file_name="dual_port_2_clock_ram.vhd",
        )
